<?php

namespace Drupal\commerce_globalonepay\Plugin\Commerce\PaymentGateway;

use Drupal\commerce_payment\CreditCard;
use Drupal\commerce_payment\Entity\PaymentInterface;
use Drupal\commerce_payment\Entity\PaymentMethodInterface;
use Drupal\commerce_payment\Exception\HardDeclineException;
use Drupal\commerce_payment\Exception\InvalidRequestException;
use Drupal\commerce_payment\Exception\PaymentGatewayException;
use Drupal\commerce_payment\PaymentMethodTypeManager;
use Drupal\commerce_payment\PaymentTypeManager;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\OnsitePaymentGatewayBase;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsRefundsInterface;
use Drupal\commerce_payment\Plugin\Commerce\PaymentGateway\SupportsVoidsInterface;
use Drupal\commerce_price\Price;
use Drupal\Component\Datetime\TimeInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Logger\LoggerChannelFactoryInterface;
use Drupal\Core\Session\AccountInterface;
use GuzzleHttp\ClientInterface;
use GuzzleHttp\Exception\RequestException;
use SimpleXMLElement;

/**
 * Provides the GlobalOnePay on-site payment gateway.
 *
 * @CommercePaymentGateway(
 *   id = "commerce_globalonepay_onsite",
 *   label = "GlobalOnePay (On-site)",
 *   display_label = "GlobalOnePay",
 *   forms = {
 *     "add-payment-method" = "Drupal\commerce_payment\PluginForm\PaymentMethodAddForm",
 *   },
 *   payment_method_types = {"credit_card"},
 *   credit_card_types = {
 *     "amex", "discover", "jcb", "maestro", "mastercard", "visa",
 *   },
 * )
 */
class GlobalOnePayOnsite extends OnsitePaymentGatewayBase implements GlobalOnePayOnsiteInterface {

  /**
   * GlobalOnePay test API URL.
   */
  const GLOBALONEPAY_API_TEST_URL = 'https://testpayments.globalone.me/merchant/xmlpayment';

  /**
   * GlobalOnePay production API URL.
   */
  const GLOBALONEPAY_API_URL = 'https://payments.globalone.me/merchant/xmlpayment';

  /**
   * The type of the terminal.
   *
   * 1 - MOTO (Mail Order/Telephone Order);
   * 2 – eCommerce;
   * 3 – Cardholder Present.
   */
  const GLOBALONEPAY_TERMINAL_TYPE = 2;

  /**
   * The type of the transaction.
   *
   * 0 – Cardholder Present (CHP) transaction;
   * 2 – Recurring;
   * 4 - MOTO (Mail Order/Telephone Order);
   * 7 – eCommerce.
   */
  const GLOBALONEPAY_TRANSACTION_TYPE = 7;

  /**
   * The logger.
   *
   * @var \Drupal\Core\Logger\LoggerChannelInterface
   */
  protected $logger;

  /**
   * The HTTP client.
   *
   * @var \GuzzleHttp\Client
   */
  protected $httpClient;

  /**
   * The current user.
   *
   * @var \Drupal\Core\Session\AccountInterface
   */
  protected $currentUser;

  /**
   * Constructs a new PaymentGatewayBase object.
   *
   * @param array $configuration
   *   A configuration array containing information about the plugin instance.
   * @param string $plugin_id
   *   The plugin_id for the plugin instance.
   * @param mixed $plugin_definition
   *   The plugin implementation definition.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\commerce_payment\PaymentTypeManager $payment_type_manager
   *   The payment type manager.
   * @param \Drupal\commerce_payment\PaymentMethodTypeManager $payment_method_type_manager
   *   The payment method type manager.
   * @param \Drupal\Component\Datetime\TimeInterface $time
   *   The time.
   * @param \Drupal\Core\Logger\LoggerChannelFactoryInterface $logger_channel_factory
   *   The logger channel factory.
   * @param \GuzzleHttp\ClientInterface $client
   *   The client.
   * @param \Drupal\Core\Session\AccountInterface $current_user
   *   The current user.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition, EntityTypeManagerInterface $entity_type_manager, PaymentTypeManager $payment_type_manager, PaymentMethodTypeManager $payment_method_type_manager, TimeInterface $time, LoggerChannelFactoryInterface $logger_channel_factory, ClientInterface $client, AccountInterface $current_user) {
    parent::__construct($configuration, $plugin_id, $plugin_definition, $entity_type_manager, $payment_type_manager, $payment_method_type_manager, $time);

    $this->logger = $logger_channel_factory->get('commerce_globalonepay');
    $this->httpClient = $client;
    $this->currentUser = $current_user;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration,
      $plugin_id,
      $plugin_definition,
      $container->get('entity_type.manager'),
      $container->get('plugin.manager.commerce_payment_type'),
      $container->get('plugin.manager.commerce_payment_method_type'),
      $container->get('datetime.time'),
      $container->get('logger.factory'),
      $container->get('http_client'),
      $container->get('current_user')
    );
  }

  /**
   * {@inheritdoc}
   */
  protected function getDefaultForms() {
    $default_forms = parent::getDefaultForms();

    if ($this instanceof SupportsRefundsInterface) {
      $default_forms['refund-payment'] = 'Drupal\commerce_globalonepay\PluginForm\RefundForm';
    }
    if ($this instanceof SupportsVoidsInterface) {
      $default_forms['void-payment'] = 'Drupal\commerce_globalonepay\PluginForm\VoidForm';
    }

    return $default_forms;
  }

  /**
   * {@inheritdoc}
   */
  public function defaultConfiguration() {
    return [
      'terminal_id' => '',
      'shared_secret' => '',
    ] + parent::defaultConfiguration();
  }

  /**
   * {@inheritdoc}
   */
  public function buildConfigurationForm(array $form, FormStateInterface $form_state) {
    $form = parent::buildConfigurationForm($form, $form_state);

    $form['terminal_id'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Terminal ID'),
      '#default_value' => $this->configuration['terminal_id'],
      '#required' => TRUE,
    ];
    $form['shared_secret'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Shared Secret'),
      '#description' => $this->t('To create new Shared Secret login to the GlobalOnePay Merchant System, navigate to "Settings" > "Terminal" and set it in "Secret" and "Confirm Secret" fields.'),
      '#default_value' => $this->configuration['shared_secret'],
      '#required' => TRUE,
    ];

    return $form;
  }

  /**
   * {@inheritdoc}
   */
  public function submitConfigurationForm(array &$form, FormStateInterface $form_state) {
    parent::submitConfigurationForm($form, $form_state);

    if (!$form_state->getErrors()) {
      $values = $form_state->getValue($form['#parents']);
      $this->configuration['terminal_id'] = $values['terminal_id'];
      $this->configuration['shared_secret'] = $values['shared_secret'];
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getApiUrl() {
    return $this->getMode() == 'test' ? self::GLOBALONEPAY_API_TEST_URL : self::GLOBALONEPAY_API_URL;
  }

  /**
   * {@inheritdoc}
   */
  public function getTerminalId() {
    return $this->configuration['terminal_id'] ?: '';
  }

  /**
   * {@inheritdoc}
   */
  public function getSharedSecret() {
    return $this->configuration['shared_secret'] ?: '';
  }

  /**
   * {@inheritdoc}
   */
  public function generateMerchantRef() {
    return $this->currentUser->id() . time() . uniqid(mt_rand(5, 15));
  }

  /**
   * {@inheritdoc}
   */
  public function generateHash(array $parameters) {
    return md5(implode('', $parameters));
  }

  /**
   * {@inheritdoc}
   */
  public function getExpirationDate(array $payment_details) {
    return $payment_details['expiration']['month'] . substr($payment_details['expiration']['year'], -2);
  }

  /**
   * {@inheritdoc}
   */
  public function getOrderId(PaymentInterface $payment) {
    return $payment->getOrderId() . '-' . date('i:s');
  }

  /**
   * {@inheritdoc}
   */
  public function setApprovalCode(PaymentInterface &$payment, $approval_code) {
    $payment->set('approval_code', $approval_code);
  }

  /**
   * {@inheritdoc}
   */
  public function prepareBody(array $parameters, SimpleXMLElement &$xml) {
    // Insert needed parameters into XML object.
    foreach ($parameters as $key => $value) {
      if (is_numeric($key)) {
        $key = 'item' . $key;
      }
      if (is_array($value)) {
        $subnode = $xml->addChild($key);
        self::prepareBody($value, $subnode);
      }
      else {
        $xml->addChild("$key", htmlspecialchars("$value"));
      }
    }
    // Return results as XML string.
    return $xml->asXML();
  }

  /**
   * {@inheritdoc}
   */
  public function prepareResult($xml_string) {
    $xml = simplexml_load_string($xml_string);
    $json = json_encode($xml);
    return json_decode($json, TRUE);
  }

  /**
   * {@inheritdoc}
   */
  public function formatResponse($data) {
    $string = '';
    foreach ($data as $key => $value) {
      $value = !empty($value) ? $value : '-';
      $string .= "$key: $value; ";
    }
    return $string;
  }

  /**
   * {@inheritdoc}
   */
  public function executeTransaction($type, array $parameters) {
    // Create new XML object.
    $xml = new \SimpleXMLElement('<?xml version="1.0" encoding="UTF-8"?>' . "<$type></$type>");

    // Add all needed parameters for new request.
    $body = $this->prepareBody($parameters, $xml);

    // Send request.
    $response = $this->httpClient->post($this->getApiUrl(), [
      'headers' => [
        'Content-Type' => 'text/xml; charset=UTF8',
        'Content-Length' => strlen($body),
      ],
      'body' => $body,
      'timeout' => 45,
    ]);

    return $this->prepareResult($response->getBody()->getContents());
  }

  /**
   * {@inheritdoc}
   */
  public function validatePayment(PaymentInterface $payment, $payment_state = 'new') {
    $this->assertPaymentState($payment, [$payment_state]);

    $payment_method = $payment->getPaymentMethod();

    switch ($payment_state) {
      case 'new':
        if (empty($payment_method)) {
          throw new HardDeclineException('The provided payment has no payment method referenced.');
        }
        elseif ($payment_method->isExpired()) {
          throw new HardDeclineException('The provided payment method has expired.');
        }

        break;

      case 'authorization':
        if ($payment->isExpired()) {
          throw new HardDeclineException('The provided transaction has expired. Authorizations are guaranteed for up to 7 days.');
        }
        if (empty($payment->getRemoteId())) {
          throw new HardDeclineException('Could not retrieve the transaction ID.');
        }
        break;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createPayment(PaymentInterface $payment, $capture = TRUE) {
    $this->validatePayment($payment, 'new');

    // Collect needed information.
    $order_id = $this->getOrderId($payment);
    $amount = $payment->getAmount();
    $datetime = date('d-m-Y:H:i:s:v');
    $hash = $this->generateHash([
      $this->getTerminalId(),
      $order_id,
      $amount->getNumber(),
      $datetime,
      $this->getSharedSecret(),
    ]);
    $payment_type = $capture ? 'PAYMENT' : 'PREAUTH';
    $payment_method = $payment->getPaymentMethod();

    // Execute the transaction.
    try {
      $data = $this->executeTransaction($payment_type, [
        'ORDERID' => $order_id,
        'TERMINALID' => $this->getTerminalId(),
        'AMOUNT' => $amount->getNumber(),
        'DATETIME' => $datetime,
        'CARDNUMBER' => $payment_method->get('card_reference')->value,
        'CARDTYPE' => $payment_method->get('card_type')->value,
        'CARDEXPIRY' => date('my', $payment_method->getExpiresTime()),
        'HASH' => $hash,
        'CURRENCY' => $amount->getCurrencyCode(),
        'TERMINALTYPE' => self::GLOBALONEPAY_TERMINAL_TYPE,
        'TRANSACTIONTYPE' => self::GLOBALONEPAY_TRANSACTION_TYPE,
      ]);

      // Check if the payment is successful.
      if (empty($data["RESPONSECODE"]) || $data["RESPONSECODE"] != 'A') {
        // Check if we have an error processing this payment.
        if (isset($data['ERRORSTRING'])) {
          throw new PaymentGatewayException("Unable to capture the payment: " . $data['ERRORSTRING']);
        }
        elseif ($data["RESPONSECODE"] == 'D' || $data["RESPONSECODE"] == 'R') {
          throw new HardDeclineException("Unable to capture the payment. Response code: " . $data["RESPONSECODE"] . "Response text: " . $data['RESPONSETEXT']);
        }
        else {
          $response_string = $this->formatResponse($data);
          throw new PaymentGatewayException('Unable to capture the payment: ' . $response_string);
        }
      }

      // Detect next step based on capture method.
      $next_state = $capture ? 'completed' : 'authorization';
      $payment->setState($next_state);

      if (!$capture) {
        // Set authorization expiration time.
        $payment->setExpiresTime($this->time->getRequestTime() + (86400 * 7));
      }

      // Store the remote ID and the approval code returned by the request.
      $this->setApprovalCode($payment, $data["APPROVALCODE"]);
      $payment->setRemoteId($data["UNIQUEREF"]);
      $payment->save();
    }
    catch (RequestException $e) {
      throw new InvalidRequestException("Unable to create a payment: " . $e->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function capturePayment(PaymentInterface $payment, Price $amount = NULL) {
    $this->validatePayment($payment, 'authorization');
    // If not specified, capture the entire amount.
    $amount = $amount ?: $payment->getAmount();
    $remote_id = $payment->getRemoteId();
    $datetime = date('d-m-Y:H:i:s:v');
    $hash = $this->generateHash([
      $this->getTerminalId(),
      $remote_id,
      $amount->getNumber(),
      $datetime,
      $this->getSharedSecret(),
    ]);

    // Execute the transaction.
    try {
      $data = $this->executeTransaction('PREAUTHCOMPLETION', [
        'UNIQUEREF' => $remote_id,
        'TERMINALID' => $this->getTerminalId(),
        'AMOUNT' => $amount->getNumber(),
        'DATETIME' => $datetime,
        'HASH' => $hash,
      ]);

      // Check if the payment is successful.
      if (empty($data["RESPONSECODE"]) || $data["RESPONSECODE"] != 'A') {
        // Check if we have an error processing this payment.
        if (isset($data['ERRORSTRING'])) {
          throw new PaymentGatewayException("Unable to capture the payment: " . $data['ERRORSTRING']);
        }
        elseif ($data["RESPONSECODE"] == 'D' && isset($data["RESPONSETEXT"])) {
          throw new PaymentGatewayException("Unable to capture the payment: " . $data['RESPONSETEXT']);
        }
        else {
          throw new PaymentGatewayException('Unable to capture the payment.');
        }
      }

      // Store the approval code returned by the request.
      $this->setApprovalCode($payment, $data["APPROVALCODE"]);

      $payment->setState('completed');
      $payment->setAmount($amount);
      $payment->save();
    }
    catch (RequestException $e) {
      throw new InvalidRequestException("Could not capture payment: " . $e->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function voidPayment(PaymentInterface $payment, $reason = NULL) {
    $this->validatePayment($payment, 'authorization');
    $remote_id = $payment->getRemoteId();

    if (empty($remote_id)) {
      throw new InvalidRequestException('Could not determine the remote payment details.');
    }

    // Collect some GlobalOnePay specific data.
    $reason = $reason ?: $this->t('Wrong payment.');
    $operator = $this->currentUser->getAccountName() ?: 'site administrator';

    // Collect the rest of needed data.
    $amount = $payment->getAmount();
    $datetime = date('d-m-Y:H:i:s:v');
    $hash = $this->generateHash([
      $this->getTerminalId(),
      $remote_id,
      $amount->getNumber(),
      $datetime,
      $this->getSharedSecret(),
    ]);

    // Execute the transaction.
    try {
      $data = $this->executeTransaction('REFUND', [
        'UNIQUEREF' => $remote_id,
        'TERMINALID' => $this->getTerminalId(),
        'AMOUNT' => $amount->getNumber(),
        'DATETIME' => $datetime,
        'HASH' => $hash,
        'OPERATOR' => $operator,
        'REASON' => $reason,
      ]);

      // Check if we have an error processing this payment.
      if (isset($data['ERRORSTRING'])) {
        throw new PaymentGatewayException("Could not void the payment: " . $data['ERRORSTRING']);
      }

      // Set correct state.
      $payment->setState('authorization_voided');
      $payment->save();
    }
    catch (RequestException $e) {
      throw new InvalidRequestException("Could not void the payment: " . $e->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function refundPayment(PaymentInterface $payment, Price $amount = NULL, $reason = NULL) {
    $this->assertPaymentState($payment, ['completed', 'partially_refunded']);
    $remote_id = $payment->getRemoteId();

    if (empty($remote_id)) {
      throw new InvalidRequestException('Could not determine the remote payment details.');
    }

    // If not specified, refund the entire amount.
    $amount = $amount ?: $payment->getAmount();
    $this->assertRefundAmount($payment, $amount);

    // Collect some GlobalOnePay specific data.
    $reason = $reason ?: $this->t('Wrong payment.');
    $operator = $this->currentUser->getAccountName() ?: $this->t('site administrator');

    // Collect the rest of needed data.
    $datetime = date('d-m-Y:H:i:s:v');
    $hash = $this->generateHash([
      $this->getTerminalId(),
      $remote_id,
      $amount->getNumber(),
      $datetime,
      $this->getSharedSecret(),
    ]);

    // Execute the transaction.
    try {
      $new_refunded_amount = $payment->getRefundedAmount()->add($amount);

      $data = $this->executeTransaction('REFUND', [
        'UNIQUEREF' => $remote_id,
        'TERMINALID' => $this->getTerminalId(),
        'AMOUNT' => $amount->getNumber(),
        'DATETIME' => $datetime,
        'HASH' => $hash,
        'OPERATOR' => $operator,
        'REASON' => $reason,
      ]);

      // Check if we have an error processing this payment.
      if (isset($data['ERRORSTRING'])) {
        throw new PaymentGatewayException("Could not refund the payment: " . $data['ERRORSTRING']);
      }

      // Set correct state.
      if ($new_refunded_amount->lessThan($payment->getAmount())) {
        $payment->setState('partially_refunded');
      }
      else {
        $payment->setState('refunded');
      }
      $payment->setRefundedAmount($new_refunded_amount);
      $payment->save();
    }
    catch (RequestException $e) {
      throw new InvalidRequestException("Could not refund the payment: " . $e->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function createPaymentMethod(PaymentMethodInterface $payment_method, array $payment_details) {
    // Check if all required data is collected.
    $required_keys = ['type', 'number', 'expiration'];
    foreach ($required_keys as $required_key) {
      if (empty($payment_details[$required_key])) {
        throw new \InvalidArgumentException(sprintf('$payment_details must contain the %s key.', $required_key));
      }
    }

    // Collect needed information.
    /** @var \Drupal\address\Plugin\Field\FieldType\AddressItem $address */
    $address = $payment_method->getBillingProfile()->get('address')->first();
    $full_name = $address->getGivenName() . ' ' . $address->getFamilyName();
    $terminal_id = $this->getTerminalId();
    $card_number = $payment_details['number'];
    $security_code = $payment_details['security_code'];
    $datetime = date('d-m-Y:H:i:s:v');
    $merchantref = $this->generateMerchantRef();
    $card_expiry = $this->getExpirationDate($payment_details);
    $card_type = $payment_details['type'];
    $hash = $this->generateHash([
      $terminal_id,
      $merchantref,
      $datetime,
      $card_number,
      $this->getExpirationDate($payment_details),
      $card_type,
      $address->getGivenName() . ' ' . $address->getFamilyName(),
      $this->getSharedSecret(),
    ]);

    // Execute the transaction.
    try {
      $data = $this->executeTransaction('SECURECARDREGISTRATION', [
        'MERCHANTREF' => $merchantref,
        'TERMINALID' => $terminal_id,
        'DATETIME' => $datetime,
        'CARDNUMBER' => $card_number,
        'CARDEXPIRY' => $card_expiry,
        'CARDTYPE' => $card_type,
        'CARDHOLDERNAME' => $full_name,
        'HASH' => $hash,
        'CVV' => $security_code,
      ]);

      // Check if we have an error processing this payment.
      if (isset($data['ERRORSTRING'])) {
        throw new HardDeclineException("Unable to verify card details: " . $data['ERRORSTRING']);
      }

      // Collect payment information into local variable.
      $payment_method->card_type = $payment_details['type'];
      // Only the last 4 numbers are safe to store.
      $payment_method->card_number = substr($payment_details['number'], -4);
      $payment_method->card_exp_month = $payment_details['expiration']['month'];
      $payment_method->card_exp_year = $payment_details['expiration']['year'];
      // Set special GlobalOnePay fields.
      $payment_method->card_reference = $data['CARDREFERENCE'];
      $payment_method->merchant_reference = $data['MERCHANTREF'];
      // Calculate expire date.
      $expires = CreditCard::calculateExpirationTimestamp($payment_details['expiration']['month'], $payment_details['expiration']['year']);
      $payment_method->setRemoteId($data['MERCHANTREF']);
      $payment_method->setExpiresTime($expires);
      $payment_method->save();
    }
    catch (RequestException $e) {
      throw new InvalidRequestException("Unable to store new secure card: " . $e->getMessage());
    }
  }

  /**
   * {@inheritdoc}
   */
  public function deletePaymentMethod(PaymentMethodInterface $payment_method) {
    // Delete the remote record.
    $terminal_id = $this->getTerminalId();
    $datetime = date('d-m-Y:H:i:s:v');
    $merchant_reference = $payment_method->get('merchant_reference')->value;
    $card_reference = $payment_method->get('card_reference')->value;
    $hash = $this->generateHash([
      $terminal_id,
      $merchant_reference,
      $datetime,
      $card_reference,
      $this->getSharedSecret(),
    ]);
    try {
      $data = $this->executeTransaction('SECURECARDREMOVAL', [
        'MERCHANTREF' => $merchant_reference,
        'CARDREFERENCE' => $card_reference,
        'TERMINALID' => $terminal_id,
        'DATETIME' => $datetime,
        'HASH' => $hash,
      ]);

      // Check if we have an error processing this payment.
      if (isset($data['ERRORSTRING'])) {
        throw new PaymentGatewayException("Unable to delete the credit card: " . $data['ERRORSTRING']);
      }
    }
    catch (RequestException $e) {
      throw new InvalidRequestException("Unable to delete the credit card: " . $e->getMessage());
    }

    // Delete the local entity.
    $payment_method->delete();
  }

}
